
import java.util.ArrayList;

import java.awt.Color;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.BoxLayout;


class Soldier {
	public Army army; // the army this soldier belongs to
	public Point2D position; // in world space coordinates, NOT in pixels
	public Point2D targetPosition; // in world space coordinates, NOT in pixels
	public Soldier enemyIAmCurrentlyAttacking = null;
	public float health = 1; // in [0,1].  0 means dead, 1 means fully healthy
	public float distanceTravelled = 0; // used to determine how to draw legs

	public Vector2D movementVector = new Vector2D( 0, 0 );

	public Soldier( Army a, Point2D initialPosition ) {
		army = a;
		position = new Point2D( initialPosition );
		targetPosition = new Point2D( initialPosition );
	}
}

class Army {
	public ArrayList< Soldier > soldiers = new ArrayList< Soldier >();
	public ArrayList< Soldier > selectedSoldiers = new ArrayList< Soldier >(); // a subset of the soldiers in this army
	public Color color;

	public Army( Color c ) {
		color = c;
	}

	public void setSelectedSoldiers( ArrayList< Soldier > setOfSoldiers ) {
		selectedSoldiers.clear();
		for ( Soldier s : setOfSoldiers ) {
			// Only add this soldier to the set of selected soldiers
			// if it belongs to this army.
			if ( s.army == this )
				selectedSoldiers.add( s );
		}
	}

	public void setTargetPositionOfSelectedSoldiers( Point2D t ) {
		for ( Soldier s : selectedSoldiers ) {
			s.targetPosition.copy( t );
		}
	}

	public ArrayList< Point2D > computePolygonEnclosingSelectedSoldiers() {
		ArrayList< Point2D > points = new ArrayList< Point2D >();
		for ( Soldier soldier : selectedSoldiers ) {
			points.add( soldier.position );
		}
		points = Point2DUtil.computeConvexHull( points );
		points = Point2DUtil.computeExpandedPolygon( points, Constant.SOLDIER_RADIUS * 2 );
		return points;
	}

}

class Game {
	public ArrayList< Army > armies = new ArrayList< Army >();

	public Game() {
		for ( int i = 0; i < Constant.NUM_ARMIES; ++i ) {
			Army army = new Army( new Color( Color.HSBtoRGB(i/(float)Constant.NUM_ARMIES,1,1) ) );
			armies.add( army );
			float angle1 = i * 2*(float)Math.PI/Constant.NUM_ARMIES;
			float radius2 = Constant.DESIRED_MIN_DISTANCE_BETWEEN_SOLDIERS * (Constant.NUM_SOLDIERS_PER_ARMY+3) / 3.0f;
			float radius1 = radius2 * (Constant.NUM_ARMIES+3) / 3.0f;
			Point2D centerOfArmy = new Point2D( radius1*(float)Math.cos(angle1), radius1*(float)Math.sin(angle1) );
			for ( int j = 0; j < Constant.NUM_SOLDIERS_PER_ARMY; ++j ) {
				float angle2 = j * 2*(float)Math.PI/Constant.NUM_SOLDIERS_PER_ARMY;
				Point2D p = new Point2D(centerOfArmy.x()+radius2*(float)Math.cos(angle2), centerOfArmy.y()+radius2*(float)Math.sin(angle2));
				army.soldiers.add( new Soldier( army, p ) );
			}
		}
	}

	public AlignedRectangle2D getBoundingRectangle() {
		AlignedRectangle2D rectangle = new AlignedRectangle2D();
		for ( Army army : armies ) {
			for ( Soldier soldier : army.soldiers ) {
				rectangle.bound( soldier.position );
			}
		}
		return rectangle;
	}

	public void advanceTimeStep() {
		// Compute the direction each soldier should move in.
		for ( Army army : armies ) {
			for ( Soldier soldier : army.soldiers ) {

				// This will store the net direction to move in.
				soldier.movementVector.copy( 0, 0 );

				// First, we check what other soldiers are nearby.
				// We don't want the soldier to get too close to any other soldiers.
				for ( Army otherArmy : armies ) {
					for ( Soldier otherSoldier : otherArmy.soldiers ) {
						if ( otherSoldier != soldier ) {
							float distanceBetweenSoldiers = soldier.position.distance( otherSoldier.position );
							if ( distanceBetweenSoldiers < Constant.DESIRED_MIN_DISTANCE_BETWEEN_SOLDIERS ) {
								Vector2D directionToMoveAwayFromOtherSoldier = Point2D.diff(soldier.position,otherSoldier.position).normalized();
								soldier.movementVector = Vector2D.sum( soldier.movementVector, directionToMoveAwayFromOtherSoldier );
							}
						}
					}
				}

				// Second, we also want the soldier to move toward its target position.
				Vector2D directionToMoveTowardTargetPosition = Point2D.diff(soldier.targetPosition,soldier.position).normalized();
				soldier.movementVector = Vector2D.sum( soldier.movementVector, directionToMoveTowardTargetPosition );

				// Adjust the length of this vector
				soldier.movementVector = Vector2D.mult( soldier.movementVector.normalized(), Constant.DISTANCE_TO_MOVE_EACH_STEP );
			}
		}

		// Now, actually move the soldiers.
		for ( Army army : armies ) {
			for ( Soldier soldier : army.soldiers ) {
				soldier.position = Point2D.sum( soldier.position, soldier.movementVector );
				soldier.distanceTravelled += soldier.movementVector.length();
			}
		}

		// Now, search the neighborhood around each soldier to see if there is an enemy to attack.
		// For each soldier, we want it to attack the *closest* enemy.
		for ( Army army : armies ) {
			for ( Soldier soldier : army.soldiers ) {
				soldier.enemyIAmCurrentlyAttacking = null;
				float distanceToEnemy = 0;
				for ( Army otherArmy : armies ) {
					for ( Soldier otherSoldier : otherArmy.soldiers ) {
						if ( otherSoldier.army != soldier.army ) {
							// The two soldiers are of different armies, therefore they are enemies.

							float distanceBetweenSoldiers = soldier.position.distance( otherSoldier.position );
							if ( distanceBetweenSoldiers < Constant.MAX_ATTACK_DISTANCE_BETWEEN_ENEMY_SOLDIERS ) {
								if (
									// Is this the first enemy we've found in the neighborhood ?
									soldier.enemyIAmCurrentlyAttacking == null
									||
									// Is this enemy closer than the previously encountered one ?
									distanceBetweenSoldiers < distanceToEnemy
								) {
									soldier.enemyIAmCurrentlyAttacking = otherSoldier;
									distanceToEnemy = distanceBetweenSoldiers;
								}
							}
						}
					}
				}

				// Now, actually attack the enemy.
				if ( soldier.enemyIAmCurrentlyAttacking != null ) {
					soldier.enemyIAmCurrentlyAttacking.health -= Constant.MAX_DAMAGE_FROM_ATTACK * Math.random() * soldier.health;
				}
			}
		}

		// Now, remove any dead soldiers
		for ( Army army : armies ) {
			for ( int indexOfSoldier = army.soldiers.size()-1; indexOfSoldier >= 0; --indexOfSoldier ) {
				Soldier s = army.soldiers.get( indexOfSoldier );
				if ( s.health <= 0 ) {
					army.soldiers.remove( indexOfSoldier );
					army.selectedSoldiers.remove( s );
				}
			}
		}

		// Allow remaining soldiers to heal by increasing their health a bit
		for ( Army army : armies ) {
			for ( int indexOfSoldier = army.soldiers.size()-1; indexOfSoldier >= 0; --indexOfSoldier ) {
				Soldier s = army.soldiers.get( indexOfSoldier );
				if ( s.health < 1 ) {
					s.health += Constant.MAX_DAMAGE_FROM_ATTACK * 0.05f;
					if ( s.health > 1 ) s.health = 1;
				}
			}
		}

	}

	public void draw( GraphicsWrapper gw ) {
		for ( Army army : armies ) {
			if ( army.selectedSoldiers != null ) { // draw polygon around selected soldiers
				ArrayList< Point2D > points = army.computePolygonEnclosingSelectedSoldiers();
				if ( points.size() > 0 ) {
					gw.setColor(army.color.getRed(),army.color.getGreen(),army.color.getBlue(),0.2f);
					gw.fillPolygon( points );
					gw.setColor(army.color);
					gw.setLineWidth(2);
					gw.drawPolygon( points );
					gw.setLineWidth(1);
				}
			}
			gw.setColor(army.color);
			for ( Soldier soldier : army.soldiers ) { // draw soldiers
				final float r0 = Constant.SOLDIER_RADIUS;
				float r = r0;
				gw.drawCircle( soldier.position.x()-r, soldier.position.y()-r, r );
				r *= soldier.health;
				gw.fillCircle( soldier.position.x()-r, soldier.position.y()-r, r );
				// draw legs
				{
					final float maxDistanceBetweenFeet = Constant.SOLDIER_RADIUS*0.9f;
					float d = soldier.distanceTravelled - 2*maxDistanceBetweenFeet*(float)Math.floor( soldier.distanceTravelled / (2*maxDistanceBetweenFeet) );
					// Now, d should be in the interval [ 0, 2*maxDistanceBetweenFeet ]
					float distanceBetweenFeet = 0;
					if ( d <= maxDistanceBetweenFeet ) distanceBetweenFeet = d;
					else distanceBetweenFeet = 2*maxDistanceBetweenFeet - d;
					final float heightOfLegs = Constant.SOLDIER_RADIUS*1.6f;
					final float lengthOfFeet = Constant.SOLDIER_RADIUS*0.3f;
					gw.drawLine(
						soldier.position.x()-distanceBetweenFeet/2+lengthOfFeet, soldier.position.y()+heightOfLegs,
						soldier.position.x()-distanceBetweenFeet/2, soldier.position.y()+heightOfLegs
					);
					gw.drawLine(
						soldier.position.x()-distanceBetweenFeet/2, soldier.position.y()+heightOfLegs,
						soldier.position.x(), soldier.position.y()
					);
					gw.drawLine(
						soldier.position.x(), soldier.position.y(),
						soldier.position.x()+distanceBetweenFeet/2, soldier.position.y()+heightOfLegs
					);
					gw.drawLine(
						soldier.position.x()+distanceBetweenFeet/2, soldier.position.y()+heightOfLegs,
						soldier.position.x()+distanceBetweenFeet/2+lengthOfFeet, soldier.position.y()+heightOfLegs
					);
				}
				if ( soldier.enemyIAmCurrentlyAttacking != null ) {
					// We want to draw a line connecting the soldier to the enemy it is attacking.
					// The line will be drawn in the color of the soldier's army, so the user
					// can see in what direction the attack is occurring.
					// We also offset the ends of the line segment so that,
					// if the enemy is also attacking this soldier, the two lines will not overlap.

					Vector2D v = Point2D.diff( soldier.enemyIAmCurrentlyAttacking.position, soldier.position );
					v = Vector2D.mult( new Vector2D( -v.y(), v.x() ).normalized(), r0*0.8f );
					gw.setLineWidth(2);
					gw.drawLine(
						soldier.position.x()+v.x(),
						soldier.position.y()+v.y(),
						soldier.enemyIAmCurrentlyAttacking.position.x()+v.x(),
						soldier.enemyIAmCurrentlyAttacking.position.y()+v.y()
					);
					gw.setLineWidth(1);
				}
			}
		}
	}

	public ArrayList< Soldier > findSoldiersInPolygon(
		ArrayList< Point2D > polygonPoints, // assumed to be in pixel coordinates
		GraphicsWrapper gw
	) {
		ArrayList< Soldier > soldiersInPolygon = new ArrayList< Soldier >();
		for ( Army army : armies ) {
			for ( Soldier soldier : army.soldiers ) {
				if ( Point2DUtil.isPointInsidePolygon( polygonPoints, gw.convertWorldSpaceUnitsToPixels(soldier.position) ) ) {
					soldiersInPolygon.add( soldier );
				}
			}
		}
		return soldiersInPolygon;
	}


}


class MyCursor {
	public int id; // identifier

	// This stores the history of positions of the "cursor" (finger)
	// in pixel coordinates.
	// The first position is where the finger pressed down,
	// and the last position is the current position of the finger.
	private ArrayList< Point2D > positions = new ArrayList< Point2D >();

	private float totalDistanceAlongDrag = 0;
	private float distanceFromStartToEndOfDrag = 0;

	public MyCursor( int id, float x, float y ) {
		this.id = id;
		positions.add( new Point2D(x,y) );
	}

	public ArrayList< Point2D > getPositions() { return positions; }

	public void addPosition( Point2D p ) {
		if ( positions.size() >= 1 ) {
			totalDistanceAlongDrag += p.distance( positions.get(positions.size()-1) );
			distanceFromStartToEndOfDrag = p.distance( positions.get(0) );
		}
		positions.add( p );
	}

	public Point2D getCurrentPosition() {
		if ( positions == null || positions.size() < 1 )
			return null;
		return positions.get( positions.size()-1 );
	}
	public Point2D getPreviousPosition() {
		if ( positions == null || positions.size() == 0 )
			return null;
		if ( positions.size() == 1 )
			return positions.get( 0 );
		return positions.get( positions.size()-2 );
	}

	public boolean doesDragLookLikeLassoGesture() {
		return totalDistanceAlongDrag / (float)distanceFromStartToEndOfDrag > 2.5f;
	}
}

class MyCursorContainer {
	private ArrayList< MyCursor > cursors = new ArrayList< MyCursor >();

	public int getNumCursors() { return cursors.size(); }
	public MyCursor getCursorByIndex( int index ) { return cursors.get( index ); }

	public int findIndexOfCursorById( int id ) {
		for ( int i = 0; i < cursors.size(); ++i ) {
			if ( cursors.get(i).id == id )
				return i;
		}
		return -1;
	}
	public MyCursor getCursorById( int id ) {
		int index = findIndexOfCursorById( id );
		return ( index == -1 ) ? null : cursors.get( index );
	}
	// returns index of updated cursor
	public int updateCursorById(
		int id,
		float x, float y
	) {
		Point2D updatedPosition = new Point2D( x, y );
		int index = findIndexOfCursorById( id );
		if ( index == -1 ) {
			cursors.add( new MyCursor( id, x, y ) );
			index = cursors.size() - 1;
		}
		MyCursor c = cursors.get( index );
		if ( ! c.getCurrentPosition().equals( updatedPosition ) ) {
			c.addPosition( updatedPosition );
		}
		return index;
	}
	public void removeCursorByIndex( int index ) {
		cursors.remove( index );
	}
	public ArrayList< Point2D > getWorldPositionsOfCursors( GraphicsWrapper gw ) {
		ArrayList< Point2D > positions = new ArrayList< Point2D >();
		for ( MyCursor cursor : cursors ) {
			positions.add( gw.convertPixelsToWorldSpaceUnits( cursor.getCurrentPosition() ) );
		}
		return positions;
	}
}



public class SimpleMultitouchRealTimeStrategyGame implements Runnable, ActionListener {

	public MultitouchFramework multitouchFramework = null;
	public GraphicsWrapper gw = null;
	JMenuItem testMenuItem1;
	JMenuItem testMenuItem2;
	JCheckBox autoZoomCheckBox;
	JCheckBox cameraModeCheckBox;
	JCheckBox pauseCheckBox;
	boolean isAutoZoomEnabled = true;
	boolean isCameraModeActive = false;
	boolean isGamePaused = false;
	JButton frameAllButton;
	JButton testButton1;
	JButton testButton2;

	Thread thread = null;
	boolean threadSuspended;

	Game game = new Game();

	MyCursorContainer cursorContainer = new MyCursorContainer();
	int mouse_x, mouse_y;

	public SimpleMultitouchRealTimeStrategyGame( MultitouchFramework mf, GraphicsWrapper gw ) {
		multitouchFramework = mf;
		this.gw = gw;
		multitouchFramework.setPreferredWindowSize(Constant.INITIAL_WINDOW_WIDTH,Constant.INITIAL_WINDOW_HEIGHT);
		gw.setFontHeight( Constant.TEXT_HEIGHT );
		gw.frame( game.getBoundingRectangle(), true );
	}

	public void actionPerformed(ActionEvent e) {
		Object source = e.getSource();
		if ( source == testMenuItem1 ) {
			System.out.println("testMenuItem1 has been selected");
		}
		else if ( source == testMenuItem2 ) {
			System.out.println("testMenuItem2 has been selected");
		}
		else if ( source == autoZoomCheckBox ) {
			isAutoZoomEnabled = ! isAutoZoomEnabled;
		}
		else if ( source == cameraModeCheckBox ) {
			isCameraModeActive = ! isCameraModeActive;
		}
		else if ( source == pauseCheckBox ) {
			isGamePaused = ! isGamePaused;
		}
		else if ( source == frameAllButton ) {
			gw.frame( game.getBoundingRectangle(), true );
		}
		else if ( source == testButton1 ) {
			System.out.println("testButton1 has been selected");
		}
		else if ( source == testButton2 ) {
			System.out.println("testButton2 has been selected");
		}
	}

	// Called by the framework when creating widgets.
	public JMenuBar createMenuBar() {
		JMenuBar menuBar = new JMenuBar();
			JMenu menu = new JMenu("File");
				testMenuItem1 = new JMenuItem("Test 1");
				testMenuItem1.addActionListener(this);
				menu.add(testMenuItem1);
				testMenuItem2 = new JMenuItem("Test 2");
				testMenuItem2.addActionListener(this);
				menu.add(testMenuItem2);
			menuBar.add(menu);
		return menuBar;
	}

	// Called by the framework when creating widgets.
	public JPanel createPanelOfWidgets() {
		JPanel panelOfWidgets = new JPanel();
		panelOfWidgets.setLayout( new BoxLayout( panelOfWidgets, BoxLayout.Y_AXIS ) );
		autoZoomCheckBox = new JCheckBox( "Auto Zoom", isAutoZoomEnabled );
		autoZoomCheckBox.addActionListener(this);
		panelOfWidgets.add( autoZoomCheckBox );
		cameraModeCheckBox = new JCheckBox( "Camera Mode", isCameraModeActive );
		cameraModeCheckBox.addActionListener(this);
		panelOfWidgets.add( cameraModeCheckBox );
		pauseCheckBox = new JCheckBox( "Pause Game", isGamePaused );
		pauseCheckBox.addActionListener(this);
		panelOfWidgets.add( pauseCheckBox );
		frameAllButton = new JButton("Frame All");
		frameAllButton.addActionListener(this);
		panelOfWidgets.add( frameAllButton );
		testButton1 = new JButton("Test number 1 ...");
		testButton1.addActionListener(this);
		panelOfWidgets.add( testButton1 );
		testButton2 = new JButton("Test number 2 ...");
		testButton2.addActionListener(this);
		panelOfWidgets.add( testButton2 );
		return panelOfWidgets;
	}

	// Called by the framework at startup time.
	public void startBackgroundWork() {
		if ( thread == null ) {
			thread = new Thread( this );
			threadSuspended = false;
			thread.start();
		}
		else {
			if ( threadSuspended ) {
				threadSuspended = false;
				synchronized( this ) {
					notify();
				}
			}
		}
	}
	public void stopBackgroundWork() {
		threadSuspended = true;
	}
	public void run() {
		try {
			int sleepIntervalInMilliseconds = 25;
			int timeSinceLastAutoZoomInMilliseconds = 0;
			while (true) {

				// Here's where the thread does some work
				synchronized( this ) {
					if ( ! isGamePaused ) {
						game.advanceTimeStep();
						if ( isAutoZoomEnabled && ! isCameraModeActive ) {
							timeSinceLastAutoZoomInMilliseconds += sleepIntervalInMilliseconds;
							if ( timeSinceLastAutoZoomInMilliseconds>=5000 && cursorContainer.getNumCursors()==0 ) {
								AlignedRectangle2D rectangle = game.getBoundingRectangle();

								// if the rectangle is very small, make it bigger
								final float minSize = 20 * Constant.SOLDIER_RADIUS;
								if ( rectangle.getDiagonal().x() < minSize || rectangle.getDiagonal().y() < minSize ) {
									Point2D p = rectangle.getCenter();
									rectangle.bound( Point2D.sum(p,new Vector2D( minSize/2,0)) );
									rectangle.bound( Point2D.sum(p,new Vector2D(-minSize/2,0)) );
									rectangle.bound( Point2D.sum(p,new Vector2D(0, minSize/2)) );
									rectangle.bound( Point2D.sum(p,new Vector2D(0,-minSize/2)) );
								}

								gw.frame( rectangle, true );
								timeSinceLastAutoZoomInMilliseconds = 0;
							}
						}
					}
				}
				multitouchFramework.requestRedraw();

				// Now the thread checks to see if it should suspend itself
				if ( threadSuspended ) {
					synchronized( this ) {
						while ( threadSuspended ) {
							wait();
						}
					}
				}
				thread.sleep( sleepIntervalInMilliseconds );  // interval given in milliseconds
			}
		}
		catch (InterruptedException e) { }
	}

	public synchronized void draw() {
		gw.clear(1,1,1);
		gw.setColor(0,0,0);
		gw.setupForDrawing();

		gw.setCoordinateSystemToWorldSpaceUnits();
		gw.enableAlphaBlending();

		// draw the armies
		game.draw( gw );

		gw.setCoordinateSystemToPixels();

		// draw cursors
		for ( int i = 0; i < cursorContainer.getNumCursors(); ++i ) {
			MyCursor cursor = cursorContainer.getCursorByIndex( i );
			gw.setColor(0,0.5f,0.5f,0.65f);
			gw.fillCircle( cursor.getCurrentPosition().x()-10, cursor.getCurrentPosition().y()-10, 10 );
			if ( ! isCameraModeActive ) {
				if ( cursor.doesDragLookLikeLassoGesture() ) {
					// draw polygon
					gw.setColor(0,0,0,0.2f);
					gw.fillPolygon( cursor.getPositions() );
				}
				else {
					// draw ink trail
					gw.setColor(0,0,0);
					gw.drawPolyline( cursor.getPositions() );
				}
			}
		}

		// draw some text
		if ( cursorContainer.getNumCursors() > 0 ) {
			gw.setFontHeight( Constant.TEXT_HEIGHT );
			String s = "[" + cursorContainer.getNumCursors() + " contacts]";
			gw.drawString(
				Constant.TEXT_HEIGHT,
				2 * Constant.TEXT_HEIGHT,
				s
			);
		}


	}

	public synchronized void keyPressed( KeyEvent e ) {
		System.out.println("Key '" + e.getKeyChar() + "' pressed!");
	}
	public synchronized void keyReleased( KeyEvent e ) {
	}
	public synchronized void keyTyped( KeyEvent e ) {
		System.out.println("Key '" + e.getKeyChar() + "' typed!");
	}
	public synchronized void mouseEntered( MouseEvent e ) {
	}
	public synchronized void mouseExited( MouseEvent e ) {
	}
	public synchronized void mouseClicked( MouseEvent e ) {
	}


	public synchronized void mousePressed(MouseEvent e) {
		mouse_x = e.getX();
		mouse_y = e.getY();

		// multitouchFramework.requestRedraw();
	}

	public synchronized void mouseReleased(MouseEvent e) {
		mouse_x = e.getX();
		mouse_y = e.getY();

		// multitouchFramework.requestRedraw();
	}


	public synchronized void mouseDragged(MouseEvent e) {

		mouse_x = e.getX();
		mouse_y = e.getY();

		// multitouchFramework.requestRedraw();
	}


	public synchronized void mouseMoved(MouseEvent e) {
		mouse_x = e.getX();
		mouse_y = e.getY();

		// multitouchFramework.requestRedraw();
	}


	public synchronized void processMultitouchInputEvent( int id, float x, float y, int type ) {
		//System.out.println("event: "+id+", "+x+", "+y+", "+type);
		int cursorIndex = cursorContainer.updateCursorById( id, x, y );

		if ( isCameraModeActive ) {
			if ( type == MultitouchFramework.TOUCH_EVENT_UP ) {
				// up event
				cursorContainer.removeCursorByIndex( cursorIndex );
			}
			else if ( type == MultitouchFramework.TOUCH_EVENT_MOVE ) {
				if ( cursorContainer.getNumCursors() == 2 ) {
					MyCursor cursor0 = cursorContainer.getCursorByIndex( 0 );
					MyCursor cursor1 = cursorContainer.getCursorByIndex( 1 );
					gw.panAndZoomBasedOnDisplacementOfTwoPoints(
						id==cursor0.id ? cursor0.getPreviousPosition() : cursor0.getCurrentPosition(),
						id==cursor1.id ? cursor1.getPreviousPosition() : cursor1.getCurrentPosition(),
						cursor0.getCurrentPosition(),
						cursor1.getCurrentPosition()
					);
				}
				else if ( cursorContainer.getNumCursors() == 1 ) {
					MyCursor cursor = cursorContainer.getCursorByIndex( 0 );
					gw.pan(
						cursor.getCurrentPosition().x() - cursor.getPreviousPosition().x(),
						cursor.getCurrentPosition().y() - cursor.getPreviousPosition().y()
					);
				}
			}
		}
		else {

			if ( type == MultitouchFramework.TOUCH_EVENT_UP ) {
				// up event

				MyCursor cursor = cursorContainer.getCursorByIndex( cursorIndex );
				if ( cursor.doesDragLookLikeLassoGesture() ) {
					ArrayList< Soldier > soldiersInPolygon = game.findSoldiersInPolygon( cursor.getPositions(), gw );

					// Count the number of soldiers there are of each army in the polygon
					int [] numSoldiersOfEachArmy = new int[ Constant.NUM_ARMIES ];
					for ( int i = 0; i < Constant.NUM_ARMIES; ++i )
						numSoldiersOfEachArmy[i] = 0;
					for ( Soldier soldier : soldiersInPolygon ) {
						numSoldiersOfEachArmy[ game.armies.indexOf(soldier.army) ] ++;
					}
					// Now find which army has the most soldiers in the polygon.
					int indexOfArmyWithMostSoldiers = 0;
					int maxNumSoldiersOfASingleArmy = 0;
					for ( int i = 0; i < Constant.NUM_ARMIES; ++i ) {
						if ( numSoldiersOfEachArmy[i] > maxNumSoldiersOfASingleArmy ) {
							indexOfArmyWithMostSoldiers = i;
							maxNumSoldiersOfASingleArmy = numSoldiersOfEachArmy[i];
						}
					}

					if ( maxNumSoldiersOfASingleArmy > 0 )
						game.armies.get(indexOfArmyWithMostSoldiers).setSelectedSoldiers( soldiersInPolygon );
				}
				else { // It wasn't a lasso gesture, it was a drag gesture.

					// Find which army's selected set the gesture started in.
					for ( Army army : game.armies ) {
						if (
							Point2DUtil.isPointInsidePolygon(
								army.computePolygonEnclosingSelectedSoldiers(),
								gw.convertPixelsToWorldSpaceUnits(cursor.getPositions().get(0))
							)
						) {
							army.setTargetPositionOfSelectedSoldiers( gw.convertPixelsToWorldSpaceUnits(cursor.getCurrentPosition()) );
						}
					}
				}


				cursorContainer.removeCursorByIndex( cursorIndex );
			}
		}


		multitouchFramework.requestRedraw();
	}




}


